use anyhow::{anyhow, bail, Context as _};
use std::{
	fmt::Write as _,
	fs::File,
	io::{self, BufWriter},
	ops::Deref,
	os::unix::process::ExitStatusExt as _,
	path::{Path, PathBuf},
	process::Command
};
use tempfile::{NamedTempFile, TempDir};

fn tempfile() -> io::Result<NamedTempFile> {
	NamedTempFile::new()
}

fn run(cmd: &[&str]) -> anyhow::Result<()> {
	let display = format!("`\"{}\"`", cmd.join("\" \""));
	eprintln!(" -> Running {display} ...");

	let status = match Command::new(cmd[0]).args(cmd.iter().skip(1)).status() {
		Ok(status) => status,
		Err(err) => bail!("Failed to run `\"{}\"`: {err}", cmd.join("\" \""))
	};
	if !status.success() {
		bail!(
			"Failed to run `\"{}\"`: Process returned exit code {}",
			cmd.join("\" \""),
			status.into_raw()
		);
	}
	Ok(())
}

fn run_output(cmd: &[&str]) -> anyhow::Result<String> {
	let display = format!("`\"{}\"`", cmd.join("\" \""));
	eprintln!(" -> Running {display} ...");

	let output = match Command::new(cmd[0]).args(cmd.iter().skip(1)).output() {
		Ok(output) => output,
		Err(err) => bail!("Failed to run {display}: {err}")
	};
	if !output.status.success() {
		bail!("Failed to run {display}: Process returned exit code {}", output.status.into_raw());
	}
	String::from_utf8(output.stdout).context("The output of {display} contained invalid characters")
}

macro_rules! run_parted {
    ($path:expr, $($arg:expr),+) => {
        run(&["parted", "-s", $path.to_str().unwrap(), $($arg,)+])
    };
}

trait ChrootCmd {
	fn into_shell_cmd(self) -> String;
}

impl<const N: usize> ChrootCmd for [&str; N] {
	fn into_shell_cmd(self) -> String {
		(&self[..]).into_shell_cmd()
	}
}

impl ChrootCmd for &[&str] {
	fn into_shell_cmd(self) -> String {
		let mut cmd = String::new();
		for arg in self {
			write!(cmd, "'{}' ", arg.replace('\'', "'\"'\"'")).unwrap();
		}
		cmd
	}
}

impl ChrootCmd for &str {
	fn into_shell_cmd(self) -> String {
		self.into()
	}
}

fn run_chroot(img: &Image, user: &str, cmd: impl ChrootCmd) -> anyhow::Result<()> {
	let cmd = format!(". /etc/profile; set -euo pipefail; {}", cmd.into_shell_cmd());
	run(&[
		"chroot",
		&img.root,
		"/usr/bin/env",
		"-i",
		"sudo",
		"-u",
		user,
		"-i",
		"-n",
		"ash",
		"-c",
		&cmd
	])
}

fn download(out: &Path, url: &str) -> anyhow::Result<()> {
	eprintln!(" -> Downloading {url} ...");
	attohttpc::get(url).send()?.write_to(BufWriter::new(File::create(out)?))?;
	Ok(())
}

pub struct Image {
	path: PathBuf,
	boot: Mount,
	dev: Mount,
	root: Mount,
	device: LoopDevice
}

impl Drop for Image {
	fn drop(&mut self) {
		eprintln!("Cleaning up ...");
	}
}

struct LoopDevice(String);

impl LoopDevice {
	fn part(&self, num: usize) -> String {
		format!("{}p{num}", self.0)
	}
}

impl Deref for LoopDevice {
	type Target = str;

	fn deref(&self) -> &Self::Target {
		&self.0
	}
}

impl Drop for LoopDevice {
	fn drop(&mut self) {
		let res = run(&["udisksctl", "loop-delete", "--no-user-interaction", "-b", self]);
		if let Err(err) = res {
			eprintln!("Failed to free loop device {}: {err}", self.0);
		}
	}
}

struct Mount(PathBuf);

impl Mount {
	fn create(device: &str) -> anyhow::Result<Self> {
		let path = TempDir::new()?.into_path();
		run(&["mount", device, &path.to_string_lossy()])?;
		Ok(Self(path))
	}
}

impl Deref for Mount {
	type Target = str;

	fn deref(&self) -> &Self::Target {
		self.0.to_str().unwrap()
	}
}

impl Drop for Mount {
	fn drop(&mut self) {
		let res = run(&["umount", self]);
		if let Err(err) = res {
			eprintln!("Failed to unmount {}: {err}", self.0.display());
		}
	}
}

mod init_first_boot;
mod init_network;
mod init_os;
mod install_packages;
mod prepare_img;
mod set_passwords;

pub use init_first_boot::init_first_boot;
pub use init_network::init_network;
pub use init_os::init_os;
pub use install_packages::{install_autostart, install_packages};
pub use prepare_img::prepare_img;
pub use set_passwords::set_passwords;
